import os
from pathlib import Path
import sys
import importlib

import pytest

from boydem import boydem, _db_manager


@pytest.fixture(autouse=True)
def run_around_tests():
    _db_manager.file_name = (
        Path(os.path.expanduser("~")) / ".boydem" / "boydem.test.pickle"
    )
    boydem.clear()
    yield
    boydem.clear()


class K:
    def __init__(self):
        self.x = 3

    def f(self):
        self.x += 1
        return self.x

    def __repr__(self):
        return f"K equals {self.x}"


def test_boydem():
    boydem.a = 3
    boydem["b"] = [1, 2, 3]
    boydem[7] = "💥"
    boydem.k = K()
    boydem.k.f()


def test_string_key():
    boydem["a"] = "b"
    assert hasattr(boydem, "a")
    assert "a" in boydem
    assert "no" not in boydem
    assert boydem["a"] == "b"
    assert boydem.a == "b"
    assert boydem.__getattr__("a") == "b"
    del boydem["a"]
    with pytest.raises(KeyError):
        hasattr(boydem, "a")


def test_int_key():
    boydem[123] = 456
    assert 123 in boydem
    assert boydem[123] == 456
    assert boydem.__getattr__(123) == 456
    del boydem[123]
    assert 123 not in boydem


def test_keys():
    boydem["a"] = 1
    boydem["b"] = 2
    boydem["c"] = 3
    assert len(boydem.keys()) == 3
    for key in boydem.keys():
        assert key in ("a", "b", "c")


def test_values():
    boydem["a"] = 1
    boydem["b"] = 2
    boydem["c"] = 3
    assert len(boydem.values()) == 3
    for value in boydem.values():
        assert value in (1, 2, 3)


def test_items():
    boydem["a"] = 1
    boydem["b"] = 2
    boydem["c"] = 3
    assert len(boydem.items()) == 3
    for key, value in boydem.items():
        assert key in ("a", "b", "c")
        assert value in (1, 2, 3)


def test_iter():
    boydem["a"] = 1
    boydem["b"] = 2
    boydem["c"] = 3
    assert len(boydem) == 3
    for key, value in boydem:
        assert key in ("a", "b", "c")
        assert value in (1, 2, 3)


def test_len():
    boydem.clear()
    boydem["a"] = "a"
    assert len(boydem) == 1
    boydem["a"] = "a"
    assert len(boydem) == 1
    boydem["b"] = "b"
    assert len(boydem) == 2
    del boydem.a
    assert len(boydem) == 1


def test_print(capsys):
    print(boydem)
    captured = capsys.readouterr()
    assert captured.out == "Nothing here 😳\n"
    boydem.dog = "Jaime"
    boydem["list"] = [1, 2, 3]
    boydem[3] = "three"
    boydem.k = K()
    boydem.k.f()
    print(boydem)
    captured = capsys.readouterr()
    assert captured.out == "dog: Jaime\nlist: [1, 2, 3]\n3: three\nk: K equals 4\n"
    assert captured.out != "dog: Jaime\nlist: [1, 2, 3]\n3: three\nk: K equals 3\n"



import os
from pathlib import Path
from typing import Any, Hashable, ItemsView, ValuesView, KeysView, Iterator, Dict

import dill


class _Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class _DBManager:
    def __init__(self) -> None:
        self.file_name = Path(os.path.expanduser("~")) / ".boydem" / "boydem.pickle"
        self.file_name.parent.mkdir(parents=True, exist_ok=True)
        self.file = None

    def set(self, obj: Any) -> None:
        with self.file_name.open("wb") as f:
            dill.dump(obj, f)

    def delete(self, obj: Any) -> None:
        self.set(obj)

    def clear(self) -> None:
        self.file_name.unlink(missing_ok=True)

    def get(self) -> Dict:
        try:
            with self.file_name.open("rb") as f:
                obj = dill.load(f)
        except FileNotFoundError:
            obj = dict()
        except EOFError:
            self.file_name.unlink()
            obj = dict()
        return obj


class _Boydem(metaclass=_Singleton):
    def __init__(self):
        self.__dict__ = _db_manager.get()

    def __setitem__(self, key: Hashable, value: Any) -> None:
        self.__setattr__(key, value)

    def __getitem__(self, item: Any) -> Any:
        return self.__getattr__(item)

    def __delitem__(self, key: Any) -> Any:
        self.__delattr__(key)

    def __setattr__(self, key: Hashable, value: Any) -> None:
        if key == "__dict__":
            super().__setattr__("__dict__", value)
        else:
            self.__dict__[key] = value
        _db_manager.set(self.__dict__)

    def __getattr__(self, item: Any) -> Any:
        return self.__dict__[item]

    def __delattr__(self, item: Any) -> None:
        del self.__dict__[item]
        _db_manager.delete(self.__dict__)

    def keys(self) -> KeysView:
        return self.__dict__.keys()

    def values(self) -> ValuesView:
        return self.__dict__.values()

    def items(self) -> ItemsView:
        return self.__dict__.items()

    def clear(self) -> None:
        self.__dict__.clear()
        _db_manager.clear()

    def __contains__(self, item: Hashable) -> bool:
        return item in self.__dict__

    def __len__(self):
        return len(self.__dict__)

    def __iter__(self) -> Iterator:
        return iter(self.items())

    def __repr__(self):
        items = self.__dict__.items()
        if items:
            res = "\n".join([f"{key}: {value}" for key, value in items])
        else:
            res = "Nothing here 😳"
        return res


_db_manager = _DBManager()
boydem = _Boydem()


__all__ = ["boydem"]
